/**
 * Copyright (C) 2001-2005 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer.common;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.speedo.generation.lib.AbstractGeneratorComponent;
import org.objectweb.speedo.lib.Personality;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;

/**
 * Abstract enhancer component. Provides methods to read and write classes to
 * or from the file system.
 *
 * Adapted from loadXXX and writeXXX methods in EnhancerTool.
 */
public abstract class AbstractEnhancerComponent extends AbstractGeneratorComponent {
    /**
     * Indicates whether the <code>.class</code> files would be found in an
     * archive.
     */
    public boolean isSrcJar;
    /**
     * System separator character (e.g '/' under Unix systems)
     */
    private final static char SEPARATOR = File.separatorChar;

    public AbstractEnhancerComponent(Personality p) {
    	super(p);
    }
    
    /**
     * Loads a specified class placed indifferently in a directory or in
     * a <code>.jar</code> archive.
     *
     * @param issrcjar indicates whether the <code>.class</code> file would be
     * found in an archive
     * @param completeName the name of the Java class to be loaded
     * @param srcFiles location of the <code>.jar</code> file or base directory
     *  of <code>.class</code> file
     * @return the JavaClass loaded
     * @exception org.objectweb.speedo.generation.enhancer.common.SpeedoEnhancerException if the file cannot be loaded
     */
    public ClassReader loadJavaClass(final boolean issrcjar,
                                        final String completeName,
                                        final String srcFiles,
                                        final boolean remove)
            throws SpeedoEnhancerException {
        ClassReader jclass = null;
        // lookup class in cache
        // if not found, parse the class from disk
        String file = completeName.replace('.', SEPARATOR) + ".class";
        if (!issrcjar) {
            file = srcFiles + SEPARATOR + file;
        }
        try {
            File f = new File(file);
            if (f.exists()) {
	            FileInputStream fis = new FileInputStream(f);
	            jclass = new ClassReader(fis);
	            fis.close();
	            if (remove) {
	                f.delete();
	            }
            } else {
                file = completeName.replace('.', '/') + ".class";
                InputStream is = getClass().getClassLoader()
                	.getResourceAsStream(file);
                if (is == null) {
                    throw new SpeedoEnhancerException(
                            "Error during loading of the class: " + file);
                }
	            jclass = new ClassReader(is);
            }
        } catch (IOException e) {
            throw new SpeedoEnhancerException("Error during loading " + file, e);
        }
        return jclass;
    }

    /**
     * Loads a specified class placed indifferently in a directory or in
     * a <code>.jar</code> archive and store it into a new directory.
     *
     * @param issrcjar indicates whether the <code>.class</code> file would be
     *  found in an archive
     * @param completeName the name of the Java class to be loaded
     * @param srcFiles location of the <code>.jar</code> file or base directory
     *  of <code>.class</code> file
     * @param storeDir location of the base directory where the java class
     * should be stored
     * @return the JavaClass loaded
     * @exception org.objectweb.speedo.generation.enhancer.common.SpeedoEnhancerException if the file cannot be loaded
     */
    public ClassReader loadJavaClass(final boolean issrcjar,
                                        final String completeName,
                                        final String srcFiles,
                                        final boolean remove,
                                        final String storeDir)
            throws SpeedoEnhancerException {
        ClassReader jclass = loadJavaClass(issrcjar, completeName, srcFiles, remove);
        ClassWriter cw = new ClassWriter(false);
        jclass.accept(cw, false);
        writeJavaClass(completeName, cw, storeDir);
        return jclass;
    }

    /**
     * Saves the new bytecode of the specified Java class under a specified base
     * directory. The file's location is created if necessary.
     *
     * @param jclass the Java class that has to be saved
     * @param srcFiles the base directory where it has to be saved
     * @exception org.objectweb.speedo.generation.enhancer.common.SpeedoEnhancerException if the file cannot be written
     */
    public void writeJavaClass(final String name,
                                  final ClassWriter jclass,
                                  final String srcFiles)
            throws SpeedoEnhancerException {
        try {
            int p = name.lastIndexOf('.');
            String pkg;
            String clas;
            if (p == -1) {
                pkg = "";
                clas = name;
            } else {
                pkg = name.substring(0, p).replace('.', SEPARATOR);
                clas = name.substring(p + 1);
            }
            File outputDir;
            if (srcFiles == null) {
                outputDir = new File(pkg);
            } else {
                outputDir = new File(srcFiles + SEPARATOR + pkg);
            }
            outputDir.mkdirs();
            File outputFile = new File(outputDir, clas + ".class");
            outputFile.createNewFile();
            FileOutputStream fos = new FileOutputStream(outputFile);
            fos.write(jclass.toByteArray());
            fos.close();
        } catch (IOException e) {
            throw new SpeedoEnhancerException("Cannot write " + name, e);
        }
    }
}
