/**
 * Copyright (C) 2001-2005 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.lib;

import org.objectweb.jorm.util.lib.StringReplace;

import java.io.File;
import java.io.Serializable;

/**
 * Represents a Speedo personality. Two well instances are defined (one for 
 * each personality): #JDO and #EJB. This class provides a parser method 
 * #getPersonality(String) retrieving the personality from its name. 
 * 
 * @author S.Chassande-Barrioz
 */
public abstract class Personality implements Serializable {
	
	public static Personality JDO;
	public static Personality EJB;  
	
	static {
		try {
			JDO = (Personality) Class.forName("org.objectweb.speedo.jdo.JDOPersonality").newInstance();
			JDO.init("jdo");
        } catch (Throwable e) {
			JDO = null;
		}
		try {
			EJB = (Personality) Class.forName("org.objectweb.speedo.ejb.EJBPersonality").newInstance();
			EJB.init("ejb");
		} catch (Throwable e) {
			EJB = null;
		}
	}

	private String name;
    private String nameUpper;
    private String classNameDot = null;
    private String classNameSlash = null;
    private String classNameUserDot = null;
    private String classNameUserSlash = null;
    private String classNameFatalDot = null;
    private String classNameFatalSlash = null;
    private String classNameDetachSlash = null;

    /**
     * Retrieve the basic runtime exception dedicated to the relevant personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public abstract RuntimeException newRuntimeException();
    public abstract RuntimeException newRuntimeException(String msg);
    public abstract RuntimeException newRuntimeException(String msg, Throwable[] nested);
    public abstract RuntimeException newRuntimeException(String msg, Throwable nested);
    public abstract RuntimeException newRuntimeException(String msg, Object failed);
    public abstract RuntimeException newRuntimeException(String msg, Throwable[] nested, Object failed);
    public abstract RuntimeException newRuntimeException(String msg, Throwable nested, Object failed);
    public abstract RuntimeException newRuntimeException(Throwable nested);

    /**
     * Retrieve the runtime user exception class dedicated to the relevant
     * personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public abstract Class getRuntimeExceptionClass();
    
    /**
     * Retrieve the class name of the runtime user exception dedicated to the
     * relevant personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public String getRuntimeExceptionClassNameDot() {
    	if (classNameDot == null) {
    		classNameDot = getRuntimeExceptionClass().getName();
    	}
    	return classNameDot;
    }
    
    /**
     * Retrieve the class name of the runtime user exception dedicated to the
     * relevant personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public String getRuntimeExceptionClassNameSlash() {
    	if (classNameSlash == null) {
    		classNameSlash = getRuntimeExceptionClass().getName().replace('.', '/');
    	}
    	return classNameSlash;
    }

    /**
     * Retrieve the runtime user exception dedicated to the relevant personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public abstract RuntimeException newUserRuntimeException();
    public abstract RuntimeException newUserRuntimeException(String msg);
    public abstract RuntimeException newUserRuntimeException(String msg, Throwable[] nested);
    public abstract RuntimeException newUserRuntimeException(String msg, Throwable nested);
    public abstract RuntimeException newUserRuntimeException(String msg, Object failed);
    public abstract RuntimeException newUserRuntimeException(String msg, Throwable[] nested, Object failed);
    public abstract RuntimeException newUserRuntimeException(String msg, Throwable nested, Object failed);
    public abstract RuntimeException newUserRuntimeException(Throwable nested);

    /**
     * Retrieve the runtime user exception class dedicated to the relevant
     * personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public abstract Class getUserRuntimeExceptionClass();
    
    /**
     * Retrieve the class name of the runtime user exception dedicated to the
     * relevant personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public String getUserRuntimeExceptionClassNameDot() {
    	if (classNameUserDot == null) {
    		classNameUserDot = getUserRuntimeExceptionClass().getName();
    	}
    	return classNameUserDot;
    }
    
    /**
     * Retrieve the class name of the runtime user exception dedicated to the
     * relevant personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public String getUserRuntimeExceptionClassNameSlash() {
    	if (classNameUserSlash == null) {
    		classNameUserSlash = getUserRuntimeExceptionClass().getName().replace('.', '/');
    	}
    	return classNameUserSlash;
    }

    /**
     * Retrieve the runtime fatal exception dedicated to the relevant personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public abstract RuntimeException newFatalRuntimeException();
    public abstract RuntimeException newFatalRuntimeException(String msg);
    public abstract RuntimeException newFatalRuntimeException(String msg, Throwable[] nested);
    public abstract RuntimeException newFatalRuntimeException(String msg, Throwable nested);
    public abstract RuntimeException newFatalRuntimeException(String msg, Object failed);
    public abstract RuntimeException newFatalRuntimeException(String msg, Throwable[] nested, Object failed);
    public abstract RuntimeException newFatalRuntimeException(String msg, Throwable nested, Object failed);
    public abstract RuntimeException newFatalRuntimeException(Throwable nested);

    /**
     * Retrieve the runtime fatal exception class dedicated to the relevant
     * personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public abstract Class getFatalRuntimeExceptionClass();
    
    /**
     * Retrieve the class name of the runtime fatal exception dedicated to the
     * relevant personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public String getFatalRuntimeExceptionClassNameDot() {
    	if (classNameFatalDot == null) {
    		classNameFatalDot = getFatalRuntimeExceptionClass().getName();
    	}
    	return classNameFatalDot;
    }
    
    /**
     * Retrieve the class name of the runtime fatal exception dedicated to the
     * relevant personality.
     * 
     * @return	The class of the user exception to be thrown.
     */
    public String getFatalRuntimeExceptionClassNameSlash() {
    	if (classNameFatalSlash == null) {
    		classNameFatalSlash = getFatalRuntimeExceptionClass().getName().replace('.', '/');
    	}
    	return classNameFatalSlash;
    }
    
    
    public abstract RuntimeException newFatalDataStoreRuntimeException();
    public abstract RuntimeException newFatalDataStoreRuntimeException(String msg);
    public abstract RuntimeException newFatalDataStoreRuntimeException(String msg, Throwable[] nested);
    public abstract RuntimeException newFatalDataStoreRuntimeException(String msg, Throwable nested);
    public abstract RuntimeException newFatalDataStoreRuntimeException(String msg, Object failed);
    public abstract RuntimeException newFatalDataStoreRuntimeException(String msg, Throwable[] nested, Object failed);
    public abstract RuntimeException newFatalDataStoreRuntimeException(String msg, Throwable nested, Object failed);
    public abstract RuntimeException newFatalDataStoreRuntimeException(Throwable nested);

	public abstract RuntimeException newDataStoreRuntimeException(String msg);
	public abstract RuntimeException newDataStoreRuntimeException(String msg, Throwable nested);
    public abstract Class getDetachedFieldAccessRuntimeExceptionClass();
    
    public String getDetachedFieldAccessExceptionClassNameSlash() {
    	if (classNameDetachSlash== null) {
    		classNameDetachSlash = getDetachedFieldAccessRuntimeExceptionClass().getName().replace('.', '/');
    	}
    	return classNameDetachSlash;
    }
    
    private void init(String n) {
    	this.name = n;
    	String nu = "";
    	for (int i = 0; i < name.length(); i++) {
    		nu += Character.toUpperCase(name.charAt(i));
    	}
    	nameUpper = nu;
    }
    
    public String getName() {
    	return name;
    }
    
    public String getPersonalityClassName(String pac, String classuffix) {
    	return pac + "." + name + "." + nameUpper + classuffix;
    }
    
    public String getGenClassName(String gcBaseName) {
		StringBuffer sb = new StringBuffer(gcBaseName);
		if (gcBaseName.endsWith(".class")) {
            // ".class" string
            final int l = sb.length();
    		sb.delete(l-6, l);
        }
        char sep = File.separatorChar;
        int idx = sb.lastIndexOf("" + sep);
        if (idx == -1) {
            sep = '.';
            idx = sb.lastIndexOf("" + sep);
            if (idx == -1) {
                throw new IllegalArgumentException("unmanaged gcBaseName: " + gcBaseName);
            }
        }
		// the sub package 'xxx' and the class prefix "XXX"
		sb.insert(idx + 1, name + sep + name.toUpperCase());
		if (sep != File.separatorChar) {
		    return StringReplace.replaceChar(
		            File.separatorChar, sep, sb.toString());
		} else {
		    return sb.toString();
		}
	}
    
    /**
     * Retrieve the PersistentObjectManager class for the associated
     * personality.
     * @return	The Java class.
     */
    public abstract Class getPOMClass();
    
    /**
     * Retrieve the PersistentObjectManagerFactory class for the associated
     * personality.
     * @return	The Java class.
     */
    public abstract Class getPOMFClass();
}
