/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.mim.api;

import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.speedo.pm.api.POManagerItf;
import org.objectweb.speedo.pm.api.POManagerFactoryItf;
import org.objectweb.speedo.query.api.QueryDefinition;
import org.objectweb.speedo.usercache.api.UserCacheManager;

import java.util.Collection;
import java.util.Map;
import java.util.Properties;



/**
 * It defines a home for a persitent class. Homes manage some properties 
 * (detacheable, cacheable, prefetch on query, prefetch on extent, ...) 
 * and some features (the user cache, named 
 * queries). It references the TPM, the PMF.
 * 
 * @author S.Chassande-Barrioz
 */
public interface HomeItf extends PClassMapping, UserCacheManager, Loggable {

    
    int PRE_NEW 	= 1; // before make persistent
    int POST_NEW 	= 2; // after make persistent
    int PRE_CREATE 	= 3; // before flushing a new persistent (INSERT)
    int POST_CREATE = 4; // after flushing a new persistent (INSERT)

    int PRE_REMOVE 	= 5; // before removing persistent an object
    int POST_REMOVE = 6; // after removing persistent an object
    int PRE_DELETE 	= 7; // before flushing a removed persistent (DELETE)
    int POST_DELETE = 8; // after flushing a removed persistent (DELETE)

    int POST_LOAD 	= 9; // after loading state into a persistent object

    int PRE_UPDATE 	= 10; // before flushing a modified persistent (UPDATE)
    int POST_UPDATE = 11; // after flushing a modified persistent (UPDATE)
    
    int PRE_CLEAR 	= 12;
    int POST_CLEAR 	= 13;
    
    int PRE_DIRTY 	= 14;
    int POST_DIRTY 	= 15;
    
    int PRE_DETACH 	= 16;
    int POST_DETACH = 17;

    int PRE_ATTACH 	= 18;
    int POST_ATTACH = 19;
	
    String getProjectName();

    /**
     * Indicates if the class has been marked as detacheable
     */
    boolean isDetachable();
    
    /**
     * Indicates if the state of persistent instance managed by this home, are 
     * kept between two transaction.
     * @return true means the state is kept between transaction.
     */
    boolean isCacheable();
    
    boolean hasToFix();
    
    boolean allLoaded();
    
    /**
     * Permit to Indicate if the state of persistent instance managed by this 
     * home, are kept between two transaction.
     */
    void setCachePolicy(byte v);
    byte NO_CACHE = 0;
    byte CACHED = 1;
    byte FIXED = 3;
    byte ALL = 11;
    
    boolean isShareable();
    
    void setShareable(boolean v);
    
    /**
     * Indicates if the locking level must be the field. A false value (default)
     * means the locking level has to be done at instance level.
     */
    boolean isFieldLockingLevel();
    
    /**
     * Permits to Indicate if the locking level must be the field. A false 
     * value (default) means the locking level has to be done at instance level.
     */
    void setFieldLockingLevel(boolean val);

    /**
	 * The versioning strategy.
	 * @see org.objectweb.speedo.metadata.SpeeedoVersion#DATE_TIME 
	 * @see org.objectweb.speedo.metadata.SpeeedoVersion#STATE_COMPARISON 
	 * @see org.objectweb.speedo.metadata.SpeeedoVersion#VERSION_NUMBER NO_VERSION
	 */
	byte getVersioningStrategy();

	/**
	 * @return the PersistenceManagerFactory representing the data support
	 * where the po is persistent.
	 */
	POManagerFactoryItf getPOManagerFactory();

	/**
	 * It assignes a PersistenceManagerFactory
	 * @param _pmf is the PersistenceManagerFactory representing the data 
	 * support where the po is persistent.
	 */
	void setPOManagerFactory(POManagerFactoryItf _pmf);


	/**
	 * @return the TransactionalPersistenceManager which manages the
	 * concurrency, loading, caching ... of the proxies.
	 */
	TransactionalPersistenceManager getTransactionalPersistenceManager();

	/**
	 * It assignes a TransactionalPersistenceManager
	 * @param _tpm is the TransactionalPersistenceManager which manage the
	 * concurrency, loading, ... of the po.
	 */
	void setTransactionalPersistenceManager(TransactionalPersistenceManager _tpm);

	/**
	 * Notifies the transactional persistency manager of a read intention for
	 * the given persistent instance.
	 * @param sp is the PersistentObjectItf which the read access is requested.
	 * @param fields the ids of the fields that may be accessed by the caller
	 * of this method. If the i-th bit of 'fields' is set to 1, then the i-th
	 * field of the given speedo accessor may be accessed by the caller of this
	 * method.
	 */
	StateItf readIntention(PersistentObjectItf sp, long[] fields);

	/**
	 * Notifies the transactional persistency manager of a write intention for
	 * the given persistent instance.
	 *
	 * @param sp is the PersistentObjectItf which the write access is requested.
	 * @param fields the ids of the fields that may be accessed by the caller
	 * of this method. If the i-th bit of 'fields' is set to 1, then the i-th
	 * field of the given speedo accessor may be accessed by the caller of this
	 * method.
	 * @param thinLock permits to allocate a lock on only a part of the 
	 * resource. The locking size is thinner than a lock on the persistent 
	 * object entirely. This parameter can be typically used to lock at the 
	 * field or the genclass element levels.
	 */
	StateItf writeIntention(PersistentObjectItf sp, long[] fields, Object thinLock);

	StateItf writeIntention(PersistentObjectItf sp, long[] fields);

	StateItf getState(PersistentObjectItf sp);

    PersistentObjectItf detachCopy(PersistentObjectItf sp, 
            POManagerItf pm, 
            Map map, 
            Object clone, 
            Collection fgHints) ;

    void attachCopy(PersistentObjectItf sp, 
        POManagerItf pm, 
        Map map, 
        Object clone, 
        StateItf sa) ;

	Properties getClassProperties();
	
	/**
	 * Defines a named query
	 * @param name is the name identifying the query
	 * @param query is the definition of the query
	 * @return the old query definition corresponding to the name, if it
	 * existed one.
	 */
	QueryDefinition addNamedQuery(String name, QueryDefinition query);

	/**
	 * Remove a named query
	 * @param name is the name of the query definition to remove
	 * @return the removed query definition (null if not found)
	 */
	QueryDefinition removeNamedQuery(String name);

	/**
	 * Retrieves a query defintion from its name. If the name is not known, the
	 * null value is returned.
	 * @param name is the name the wanted query definition
	 */
	QueryDefinition getNamedQuery(String name);
	
		
	void initSH();
	
	String getPath();
	
	/**
	 * Indicates if the prefetching must be activated on queries when they 
	 * return instance corresponding to this home.
	 */
	void setPrefetchOnQuery(boolean prefetch);
	
	boolean getPrefetchOnQuery();
	
	/**
	 * Indicates if the prefetching must be activated on extents when they 
	 * return instance corresponding to this home.
	 */
	void setPrefetchOnExtent(boolean prefetch);
	
	boolean getPrefetchOnExtent();
	
	/**
	 * Indicates if the prefetching must be activated genclass loading. This
	 * method is only applicable for gen class home.
	 */
	void setPrefetchOnGenClass(boolean prefetch);

	boolean getPrefetchOnGenClass();
	
	void addInstanceLifeCycleListener(Object listener);

	void removeInstanceLifeCycleListener(Object listener);
	
    void sendEvent(int eventType, Object source, Object target);
    
    boolean hasInstanceLifeCycleListeners();
}
