/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */
package org.objectweb.speedo.pm.api;

import org.objectweb.jorm.naming.api.PName;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.concurrency.lib.Semaphore;
import org.objectweb.speedo.mim.api.FetchPlanItf;
import org.objectweb.speedo.mim.api.PersistentObjectItf;
import org.objectweb.speedo.workingset.api.TransactionItf;

import java.util.Collection;
import java.util.Map;

import javax.transaction.Synchronization;

/**
 * defines a manager of persistent instance. 
 * A POManagerItf is a javax.transaction.Synchronization for the JTA
 * integration. This permits to support the container transaction demarcation.
 * On the beforeCompletion method the perseus transaction is prepared.
 * A po manager used a TransactionalPersistenceManager for managing the
 * concurrency, the loading and the caching aspects.
 *
 * @see org.objectweb.speedo.pm.api.POManagerFactoryItf
 * @see org.objectweb.speedo.pm.api.POManagerSwitchItf
 * @see org.objectweb.perseus.persistence.api.TransactionalPersistenceManager
 * 
 * @author S.Chassande-Barrioz
 */
public interface POManagerItf extends Synchronization {

    /**
	 * Retrieves the TransactionalPersistenceManager used by this POManagerItf.
	 */
	TransactionalPersistenceManager getTransactionalPersistenceManager();

    /**
     * Opens the persistent manager. This operation is the opposite of the
	 * javax.jdo.PersistenceManager.close() method. It prepares a POManagerItf
	 * to be used. During the preparation, the optimistic and multithread modes
	 * are initialized.
     * @param connectionSpec is the information to access to the data store
     * (user, password, ...)
     */
    void open(Object connectionSpec);

	/**
	 * @return the connection information to access the data store
	 */
    Object getConnectionSpec();

    /**
     * Signal to the persistence maneger that it is used. A persistence managed
	 * can be used by several thread. In this case each thread have done a
	 * PersistenceManagerFactory.getPersistentceManager() to obtain a po
	 * manager instance. The threads will do a close() operation, but only the
	 * last has to be taken in account. This method permits to knwon how many
	 * users uses the current pomanager.
     */
    void addUse();

	/**
	 * Retrieves the semaphore object permiting the multithreading mode.
	 */
    Semaphore getSemaphore();
    
    /**
     * @return the unique transaction associate to the POM.
     */
    TransactionItf getSpeedoTransaction();
    
    /**
     * Close the POM
     */
    void closePOManager();
    
    /**
     * @return true if the POM is closed, otherwise false.
     */
    boolean isPOMClosed();
    
    /**
     * @return the factory of this POM.
     */
    POManagerFactoryItf getPOManagerFactory();
    
    /**
     * Encodes a PName into a serializable representation which could be decoded
     * later. The representation depends on the PName structure.
     * @param po is a persistent object
     */
    Object getEncodedPName(PersistentObjectItf po);
    
    PName decodeIdentifier(Class aClass, Object s);
    
    /**
     * Make persistent a PersistentObjectItf. This method does the same thing than the
     * PersistenceManager.makePersistent(Object) method except the call to the
     * bind to the POManagerItf to the current thread.
     * @param sp is the instance to make persistent.
     * @param map is the context of the attachment process if any. If no attachment, it is null.
     *
     * @return TODO
     */
    Object speedoMakePersistent(PersistentObjectItf sp, Map map);

    /**
     * Delete persistent a PersistentObjectItf. This method does the same thing than the
     * PersistenceManager.deletePersistent(Object) method except the call to the
     * bind to the POManagerItf to the current thread.
     *
     * @param o is the instance to make persistent.
     */
    void speedoDeletePersistent(Object o);

    void speedoDeletePersistentAll(Object[] o);
    
    void speedoDeletePersistent(Object oid, Class pc);

    Object speedoGetObject(PName pn, boolean validate);
    
    Object speedoDetachCopy(PersistentObjectItf sp, Map map, Collection fgHints);
    
    Object speedoAttachCopy(Object detached, Map map);
    
    void speedoRefresh(PersistentObjectItf sp, Map map, Collection fgHints);
    
    void speedoRetrieve(PersistentObjectItf sp, Map map, Collection fgHints);
    
    FetchPlanItf speedoGetFetchPlan();
}
