/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.usercache.api;

/**
 * Defines an user cache of persistent objects. The user cache maintains 
 * bindings between an object identifier and the user key. This cache can be
 * implemented such as a coherent double map (oid=>key and key=>oid). 
 * The user cache entry can be removed (unbound) from the user key or from the
 * object identifier. Indeed when an object is removed from the real cache, the
 * persistent object is no more availlable, its user key too.
 * 
 * @author S.Chassande-Barrioz
 */
public interface UserCache {

	/**
	 * Look for an identifier of persistent instance from an user key 
	 * @param key is the user key of the searched persistent instance (not null)
	 * @return the identifier of the persistent instance corresponding to the
	 * user key. If not found, a null value is returned.
	 */
	Object lookup(Object key);

	/**
	 * Adds an entry into user cache. If an entry already exists with the same
	 * identifier or the same key, it has been replaced. 
	 * 
	 * @param key is the user key the user of the persistent instance (not null)
	 * @param oid is the identifier of the persistent instance (not null)
	 * 
	 * @return the previous associated identifier of a persistent instance 
	 * associated to the specified user key. The null value is returned if no 
	 * entry was previously bound.
	 */
	Object bind(Object key, Object oid);
		
	/**
	 * Forces the eviction of an entry from an user key
	 * @param key is the user key of a persistent instance (not null)
	 * @return the object identifier if found, otherwise null
	 */
	Object unbindFromKey(Object key);

	/**
	 * Forces the eviction of an entry from an object identifier
	 * @param oid is the identifier of a persistent instance (not null)
	 * @return the user key if found, otherwise null
	 */
	Object unbindFromOID(Object oid);
	
    String getName();
    
    int getId();
    
    boolean isActive();
    
    String[] getIndexFieldNames();
	
}
